"""Runs the development web interface for ADMET-AI (Flask)."""
import random
import string
from datetime import timedelta
from threading import Thread

import matplotlib
from tap import tapify

from admet_ai.admet_info import load_admet_info
from admet_ai.drugbank import load_drugbank
from admet_ai.web.app import app
from admet_ai.web.app.models import load_admet_model
from admet_ai.web.app.storage import cleanup_storage

matplotlib.use("Agg")


def setup_web(
    secret_key: str = "".join(
        random.choices(string.ascii_letters + string.digits, k=20)
    ),
    session_lifetime: int = 5 * 60,
    heartbeat_frequency: int = 60,
    max_molecules: int = 1000,
    max_visible_molecules: int = 25,
    no_cache_molecules: bool = False,
) -> None:
    """Sets up the ADMET-AI website.

    :param secret_key: Secret key for Flask app.
    :param session_lifetime: Session lifetime in seconds.
    :param heartbeat_frequency: Frequency of client heartbeat in seconds.
    :param max_molecules: Maximum number of molecules to allow predictions for.
    :param max_visible_molecules: Maximum number of molecules to display.
    :param no_cache_molecules: Whether to turn off molecule caching (reduces memory but slows down predictions).
    """
    # Set up Flask app variables
    app.secret_key = secret_key
    app.permanent_session_lifetime = timedelta(seconds=session_lifetime)
    app.config["SESSION_LIFETIME"] = session_lifetime
    app.config["HEARTBEAT_FREQUENCY"] = heartbeat_frequency
    app.config["MAX_MOLECULES"] = max_molecules
    app.config["MAX_VISIBLE_MOLECULES"] = max_visible_molecules
    app.config["CACHE_MOLECULES"] = not no_cache_molecules

    # Load ADMET info, DrugBank, and models into memory
    with app.app_context():
        print("Loading ADMET INFO")
        load_admet_info()

        print("Loading DrugBank")
        load_drugbank()

        print("Loading models")
        load_admet_model()

    # Set up garbage collection
    Thread(target=cleanup_storage).start()


def admet_web(host: str = "127.0.0.1", port: int = 5000) -> None:
    """Runs the ADMET-AI website locally.

    :param host: Host to run the website on.
    :param port: Port to run the website on.
    """
    # Set up web app with command line arguments
    tapify(setup_web)

    # Run web app
    app.run(host=host, port=port)



"""Contains global variables for use in the ADMET-AI website."""
import time

import pandas as pd

from admet_ai.web.app import app

USER_TO_PREDS: dict[str, pd.DataFrame] = {}
USER_TO_LAST_ACTIVITY: dict[str, float] = {}


def get_user_preds(user_id: str) -> pd.DataFrame:
    """Gets the user's predictions."""
    return USER_TO_PREDS.get(user_id, pd.DataFrame())


def set_user_preds(user_id: str, preds_df: pd.DataFrame) -> None:
    """Sets the user's predictions."""
    USER_TO_PREDS[user_id] = preds_df


def update_user_activity(user_id: str) -> None:
    """Updates the user's last activity time."""
    USER_TO_LAST_ACTIVITY[user_id] = time.time()


def cleanup_storage() -> None:
    """Clean up storage by removing data from users that are no longer active."""
    print("Starting cleanup")

    while True:
        # Wait for cleanup_frequency minutes
        cleanup_frequency = app.config["SESSION_LIFETIME"]
        time.sleep(cleanup_frequency)

        # Remove data from users that have been inactive
        num_removed = 0
        now = time.time()
        for user_id, last_activity in list(USER_TO_LAST_ACTIVITY.items()):
            if now - last_activity > cleanup_frequency:
                del USER_TO_PREDS[user_id]
                del USER_TO_LAST_ACTIVITY[user_id]
                num_removed += 1

        print(
            f"Cleanup removed data from {num_removed:,} users with {len(USER_TO_PREDS):,} users remaining."
        )



"""Defines functions for ADMET-AI models."""
from admet_ai import ADMETModel
from admet_ai.web.app import app


ADMET_MODEL: ADMETModel | None = None


def load_admet_model() -> None:
    """Loads the models into memory."""
    global ADMET_MODEL

    ADMET_MODEL = ADMETModel(
        models_dir=app.config["MODELS_DIR"],
        drugbank_path=app.config["DRUGBANK_PATH"],
        num_workers=app.config["NUM_WORKERS"],
        cache_molecules=app.config["CACHE_MOLECULES"],
    )


def get_admet_model() -> ADMETModel:
    """Get the ADMET-AI model.

    :return: The ADMET-AI model.
    """
    return ADMET_MODEL



"""Compute physicochemical properties using RDKit."""
import pandas as pd
from rdkit import Chem
from rdkit.Chem.Crippen import MolLogP
from rdkit.Chem.Descriptors import MolWt
from rdkit.Chem.QED import qed
from rdkit.Chem.rdMolDescriptors import (
    CalcNumAtomStereoCenters,
    CalcNumHBA,
    CalcNumHBD,
    CalcTPSA,
)
from tqdm import tqdm


def lipinski_rule_of_five(mol: Chem.Mol) -> float:
    """Determines how many of the Lipinski rules are satisfied by the molecule.

    :param mol: An RDKit molecule.
    :return: The number of Lipinski rules satisfied by the molecule.
    """
    return float(
        sum(
            [
                MolWt(mol) <= 500,
                MolLogP(mol) <= 5,
                CalcNumHBA(mol) <= 10,
                CalcNumHBD(mol) <= 5,
            ]
        )
    )


PHYSCHEM_PROPERTY_TO_FUNCTION = {
    "molecular_weight": MolWt,
    "logP": MolLogP,
    "hydrogen_bond_acceptors": CalcNumHBA,
    "hydrogen_bond_donors": CalcNumHBD,
    "Lipinski": lipinski_rule_of_five,
    "QED": qed,
    "stereo_centers": CalcNumAtomStereoCenters,
    "tpsa": CalcTPSA,
}


def compute_physicochemical_properties(
    all_smiles: list[str], mols: list[Chem.Mol] | None = None
) -> pd.DataFrame:
    """Compute physicochemical properties for a list of molecules.

    :param all_smiles: A list of SMILES.
    :param mols: A list of RDKit molecules. If None, RDKit molecules will be computed from the SMILES.
    :return: A DataFrame containing the computed physicochemical properties with SMILES strings as the index.
    """
    # Compute RDKit molecules if needed
    if mols is None:
        mols = [Chem.MolFromSmiles(smiles) for smiles in all_smiles]
    else:
        assert len(all_smiles) == len(mols)

    # Compute phyiscochemical properties and put in DataFrame with SMILES as index
    physchem_properties = pd.DataFrame(
        data=[
            {
                property_name: property_function(mol)
                for property_name, property_function in PHYSCHEM_PROPERTY_TO_FUNCTION.items()
            }
            for mol in tqdm(mols, desc="Computing physchem properties")
        ],
        index=all_smiles,
    )

    return physchem_properties



"""Contains constants used throughout ADMET-AI."""
from importlib import resources


# Paths to data and models
with resources.path("admet_ai", "resources") as resources_dir:
    DEFAULT_ADMET_PATH = resources_dir / "data" / "admet.csv"
    DEFAULT_DRUGBANK_PATH = resources_dir / "data" / "drugbank_approved.csv"
    DEFAULT_MODELS_DIR = resources_dir / "models"

# DrugBank columns
DRUGBANK_ID_COLUMN = "id"
DRUGBANK_NAME_COLUMN = "name"
DRUGBANK_SMILES_COLUMN = "smiles"
DRUGBANK_ATC_PREFIX = "atc"
DRUGBANK_ATC_NAME_PREFIX = "atc_name"
DRUGBANK_ATC_CODE_COLUMN = DRUGBANK_ATC_PREFIX
DRUGBANK_DELIMITER = ";"



"""Utility functions related to the ADMET-AI web server."""
import re
from pathlib import Path
from tempfile import TemporaryDirectory

import pandas as pd
from chemprop.data.data import SMILES_TO_MOL
from flask import request
from rdkit import Chem
from werkzeug.utils import secure_filename

from admet_ai.web.app import app


SVG_WIDTH_PATTERN = re.compile(r"width=['\"]\d+(\.\d+)?[a-z]+['\"]")
SVG_HEIGHT_PATTERN = re.compile(r"height=['\"]\d+(\.\d+)?[a-z]+['\"]")


def get_smiles_from_request() -> tuple[list[str] | None, str | None]:
    """Gets SMILES from a request.

    :return: A tuple with a list of SMILES or None and an error message or None.
    """
    # Get SMILES from request
    if request.form["text-smiles"] != "":
        smiles = request.form["text-smiles"].split("\n")
    elif request.form["draw-smiles"] != "":
        smiles = [request.form["draw-smiles"]]
    else:
        # Upload data file with SMILES
        data = request.files["data"]
        data_name = secure_filename(data.filename)
        smiles_column = request.form["smiles-column"]

        with TemporaryDirectory() as temp_dir:
            data_path = str(Path(temp_dir) / data_name)
            data.save(data_path)
            df = pd.read_csv(data_path)

            if smiles_column in df:
                smiles = df[smiles_column].astype(str).tolist()
            else:
                return None, f"SMILES column '{smiles_column}' not found in data file."

    # Strip SMILES of whitespace
    smiles = [smile.strip() for smile in smiles]

    # Skip empty lines
    smiles = [smile for smile in smiles if smile != ""]

    return smiles, None


def smiles_to_mols(smiles: list[str]) -> list[Chem.Mol]:
    """Convert a list of SMILES to a list of RDKit molecules with caching if turned on.

    :param smiles: A list of SMILES.
    :return: A list of RDKit molecules.
    """
    mols = []
    for smile in smiles:
        if smile in SMILES_TO_MOL:
            mol = SMILES_TO_MOL[smile]
        else:
            mol = Chem.MolFromSmiles(smile)

        mols.append(mol)

        if app.config["CACHE_MOLECULES"]:
            SMILES_TO_MOL[smile] = mol

    return mols


def string_to_html_sup(string: str) -> str:
    """Converts a string with an exponential to HTML superscript.

    :param string: A string.
    :return: The string with an exponential in HTML superscript.
    """
    return re.sub(r"\^(-?\d+)", r"<sup>\1</sup>", string)


def replace_svg_dimensions(svg_content: str) -> str:
    """Replace the SVG width and height with 100%.

    :param svg_content: The SVG content.
    :return: The SVG content with the width and height replaced with 100%.
    """
    # Replacing the width and height with 100%
    svg_content = SVG_WIDTH_PATTERN.sub('width="100%"', svg_content)
    svg_content = SVG_HEIGHT_PATTERN.sub('height="100%"', svg_content)

    return svg_content



"""Defines functions for ADMET-AI plots."""
import re
from io import BytesIO

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from rdkit import Chem
from rdkit.Chem.Draw.rdMolDraw2D import MolDraw2DSVG

from admet_ai.admet_info import (
    get_admet_id_to_units,
    get_admet_name_to_id,
)


def string_to_latex_sup(string: str) -> str:
    """Converts a string with an exponential to LaTeX superscript.

    :param string: A string.
    :return: The string with an exponential in LaTeX superscript.
    """
    return re.sub(r"\^(\d+)", r"$^{\1}$", string)


def plot_drugbank_reference(
    preds_df: pd.DataFrame,
    drugbank_df: pd.DataFrame,
    x_property_name: str | None = None,
    y_property_name: str | None = None,
    max_molecule_num: int | None = None,
    image_type: str = "svg",
) -> bytes:
    """Creates a 2D scatter plot of the DrugBank reference set vs the new set of molecules on two properties.

    :param preds_df: A DataFrame containing the predictions on the new molecules.
    :param drugbank_df: A DataFrame containing the DrugBank reference set.
    :param x_property_name: The name of the property to plot on the x-axis.
    :param y_property_name: The name of the property to plot on the y-axis.
    :param max_molecule_num: If provided, will display molecule numbers up to this number.
    :param image_type: The image type for the plot (e.g., svg).
    :return: Bytes containing the plot.
    """
    # Set default values
    if x_property_name is None:
        x_property_name = "Human Intestinal Absorption"

    if y_property_name is None:
        y_property_name = "Clinical Toxicity"

    # Map property names to IDs
    admet_name_to_id = get_admet_name_to_id()
    x_property_id = admet_name_to_id[x_property_name]
    y_property_id = admet_name_to_id[y_property_name]

    # Scatter plot of DrugBank molecules with histogram marginals
    sns.jointplot(
        x=drugbank_df[x_property_id],
        y=drugbank_df[y_property_id],
        kind="scatter",
        marginal_kws=dict(bins=50, fill=True),
        label="DrugBank Reference",
    )

    # Scatter plot of new molecules
    if len(preds_df) > 0:
        sns.scatterplot(
            x=preds_df[x_property_id],
            y=preds_df[y_property_id],
            color="red",
            marker="*",
            s=200,
            label="Input Molecule" + ("s" if len(preds_df) > 1 else ""),
        )

    # Add molecule numbers
    if max_molecule_num is not None:
        # Create buffer around scatter plot points
        x_min, x_max = plt.xlim()
        x_range = x_max - x_min
        x_buffer = x_range * 0.01

        y_min, y_max = plt.ylim()
        y_range = y_max - y_min
        y_buffer = y_range * 0.01

        # Get x and y values for first max_molecule_num molecules
        x_vals = preds_df[x_property_id].values[:max_molecule_num]
        y_vals = preds_df[y_property_id].values[:max_molecule_num]

        # Add molecule numbers for first max_molecule_num molecules
        for i, (x_val, y_val) in enumerate(zip(x_vals, y_vals)):
            plt.text(
                x_val - x_buffer,
                y_val + y_buffer,
                str(i + 1),
                horizontalalignment="right",
                verticalalignment="bottom",
                c="red",
            )

    # Get ADMET property units
    admet_id_to_units = get_admet_id_to_units()

    x_property_units = admet_id_to_units[x_property_id]
    y_property_units = admet_id_to_units[y_property_id]

    x_property_units = x_property_units if x_property_units != "-" else "probability"
    y_property_units = y_property_units if y_property_units != "-" else "probability"

    # Set axis labels
    plt.xlabel(f"{x_property_name} ({string_to_latex_sup(x_property_units)})")
    plt.ylabel(f"{y_property_name} ({string_to_latex_sup(y_property_units)})")

    # Save plot as svg to pass to frontend
    buf = BytesIO()
    plt.savefig(buf, format=image_type, bbox_inches="tight")
    plt.close()
    buf.seek(0)
    plot = buf.getvalue()

    return plot


def plot_radial_summary(
    property_id_to_percentile: dict[str, float],
    percentile_suffix: str = "",
    image_type: str = "svg",
) -> bytes:
    """Creates a radial plot summary of important properties of a molecule in terms of DrugBank approved percentiles.

    :param property_id_to_percentile: A dictionary mapping property IDs to their DrugBank approved percentiles.
                                      Keys are the property name along with the percentile_suffix.
    :param percentile_suffix: The suffix to add to the property names to get the DrugBank approved percentiles.
    :param image_type: The image type for the plot (e.g., svg).
    :return: Bytes containing the plot.
    """
    # Set max percentile
    max_percentile = 100

    # Set up properties
    properties = {
        "Blood-Brain Barrier Safe": {
            "percentile": max_percentile
            - property_id_to_percentile[f"BBB_Martins_{percentile_suffix}"],
        },
        "Non-\nToxic": {
            "percentile": max_percentile
            - property_id_to_percentile[f"ClinTox_{percentile_suffix}"],
            "vertical_alignment": "bottom",
        },
        "Soluble": {
            "percentile": property_id_to_percentile[
                f"Solubility_AqSolDB_{percentile_suffix}"
            ],
            "vertical_alignment": "top",
        },
        "Bioavailable": {
            "percentile": property_id_to_percentile[
                f"Bioavailability_Ma_{percentile_suffix}"
            ],
            "vertical_alignment": "top",
        },
        "hERG\nSafe": {
            "percentile": max_percentile
            - property_id_to_percentile[f"hERG_{percentile_suffix}"],
            "vertical_alignment": "bottom",
        },
    }
    property_names = [property_name for property_name in properties]
    percentiles = [
        properties[property_name]["percentile"] for property_name in properties
    ]

    # Calculate the angles of the plot (angles start at pi / 2 and go counter-clockwise)
    angles = (
        (np.linspace(0, 2 * np.pi, len(properties), endpoint=False) + np.pi / 2)
        % (2 * np.pi)
    ).tolist()

    # Complete the loop
    percentiles += percentiles[:1]
    angles += angles[:1]

    # Create a plot
    fig, ax = plt.subplots(figsize=(3, 3), subplot_kw=dict(polar=True))

    # Plot the data
    ax.fill(angles, percentiles, color="red", alpha=0.25)
    ax.plot(angles, percentiles, color="red", linewidth=2)

    # Set y limits
    ax.set_ylim(0, 100)

    # Labels for radial lines
    yticks = [0, 25, 50, 75, 100]
    yticklabels = [str(ytick) for ytick in yticks]
    ax.set_yticks(yticks)
    ax.set_yticklabels(yticklabels)
    ax.set_rlabel_position(335)

    # Labels for categories
    ax.set_xticks(angles[:-1])
    ax.set_xticklabels(property_names)

    # Adjust xticklabels so they don't overlap the plot
    for label, property_name in zip(ax.get_xticklabels(), property_names):
        if "vertical_alignment" in properties[property_name]:
            label.set_verticalalignment(properties[property_name]["vertical_alignment"])

    # Make the plot square (to match square molecule images)
    ax.set_aspect("equal", "box")

    # Ensure no text labels are cut off
    plt.tight_layout()

    # Save plot
    buf = BytesIO()
    plt.savefig(buf, format=image_type)
    plt.close()
    buf.seek(0)
    plot = buf.getvalue()

    return plot


def plot_molecule_svg(mol: str | Chem.Mol) -> str:
    """Plots a molecule as an SVG image.

    :param mol: A SMILES string or RDKit molecule.
    :return: An SVG image of the molecule.
    """
    # Convert SMILES to Mol if needed
    if isinstance(mol, str):
        mol = Chem.MolFromSmiles(mol)

    # Convert Mol to SVG
    d = MolDraw2DSVG(200, 200)
    d.DrawMolecule(mol)
    d.FinishDrawing()
    smiles_svg = d.GetDrawingText()

    return smiles_svg



"""Make predictions on a dataset using Chemprop-RDKit models trained on TDC ADMET data."""
from pathlib import Path

import pandas as pd
from tap import tapify

from admet_ai import ADMETModel
from admet_ai.constants import DEFAULT_DRUGBANK_PATH, DEFAULT_MODELS_DIR
from admet_ai.utils import load_and_preprocess_data


def admet_predict(
    data_path: Path,
    models_dir: Path = DEFAULT_MODELS_DIR,
    save_path: Path | None = None,
    include_physchem: bool = True,
    drugbank_path: Path | None = DEFAULT_DRUGBANK_PATH,
    atc_code: str | None = None,
    smiles_column: str = "smiles",
    num_workers: int | None = None,
    no_cache_molecules: bool = True,
) -> None:
    """Make predictions on a dataset using Chemprop-RDKit models trained on TDC ADMET data.

    :param data_path: Path to a CSV file containing a dataset of molecules.
    :param models_dir: Path to a directory containing subdirectories, each of which contains an ensemble
                       of Chemprop-RDKit models.
    :param save_path: Path to a CSV file where predictions will be saved. If None, defaults to data_path.
    :param include_physchem: Whether to include physicochemical properties in the predictions.
    :param drugbank_path: Path to a CSV file containing DrugBank approved molecules
                          with ADMET predictions and ATC codes.
    :param atc_code: The ATC code to filter the DrugBank reference set by.
                     If None, the entire DrugBank reference set will be used.
    :param smiles_column: Name of the column containing SMILES strings.
    :param num_workers: Number of workers for the data loader. Zero workers (i.e., sequential data loading)
                        may be faster if not using a GPU, while multiple workers (e.g., 8) are faster with a GPU.
                        If None, defaults to 0 if no GPU is available and 8 if a GPU is available.
    :param no_cache_molecules: Whether to not cache molecules.
                               Caching improves prediction speed but requires more memory.

    """
    # Load and preprocess data
    data = load_and_preprocess_data(data_path=data_path, smiles_column=smiles_column)

    # Build ADMETModel
    model = ADMETModel(
        models_dir=models_dir,
        include_physchem=include_physchem,
        drugbank_path=drugbank_path,
        atc_code=atc_code,
        num_workers=num_workers,
        cache_molecules=not no_cache_molecules,
    )

    # Make predictions
    preds = model.predict(smiles=list(data.index))

    # Merge data and preds
    data_with_preds = pd.concat((data, preds), axis=1)

    # Save predictions
    if save_path is None:
        save_path = data_path

    save_path.parent.mkdir(parents=True, exist_ok=True)
    data_with_preds.to_csv(save_path, index_label=smiles_column)


def admet_predict_command_line() -> None:
    """Run admet_predict from the command line."""
    tapify(admet_predict)



"""Defines functions for ADMET info."""
from pathlib import Path

import pandas as pd

from admet_ai.constants import DEFAULT_ADMET_PATH


ADMET_DF = pd.DataFrame()
ADMET_ID_TO_NAME: dict[str, str] = {}
ADMET_NAME_TO_ID: dict[str, str] = {}
ADMET_ID_TO_UNITS: dict[str, str] = {}


def load_admet_info(admet_path: Path = DEFAULT_ADMET_PATH) -> None:
    """Loads the ADMET info."""
    # Set up global variables
    global ADMET_DF, ADMET_ID_TO_NAME, ADMET_ID_TO_UNITS, ADMET_NAME_TO_ID

    # Load ADMET info DataFrame
    ADMET_DF = pd.read_csv(admet_path)

    # Map ADMET IDs to names and vice versa
    ADMET_ID_TO_NAME = dict(zip(ADMET_DF["id"], ADMET_DF["name"]))
    ADMET_NAME_TO_ID = dict(zip(ADMET_DF["name"], ADMET_DF["id"]))

    # Map ADMET IDs to units
    ADMET_ID_TO_UNITS = dict(zip(ADMET_DF["id"], ADMET_DF["units"]))


def lazy_load_admet_info(func: callable) -> callable:
    """Decorator to lazily load the ADMET info."""

    def wrapper(*args, **kwargs):
        if ADMET_DF.empty:
            load_admet_info()

        return func(*args, **kwargs)

    return wrapper


@lazy_load_admet_info
def get_admet_info() -> pd.DataFrame:
    """Get the ADMET info.

    :return: A DataFrame containing the ADMET info.
    """
    return ADMET_DF


@lazy_load_admet_info
def get_admet_id_to_name() -> dict[str, str]:
    """Get the ADMET ID to name mapping.

    :return: A dictionary mapping ADMET IDs to names.
    """
    return ADMET_ID_TO_NAME


@lazy_load_admet_info
def get_admet_name_to_id() -> dict[str, str]:
    """Get the ADMET name to ID mapping.

    :return: A dictionary mapping ADMET names to IDs.
    """
    return ADMET_NAME_TO_ID


@lazy_load_admet_info
def get_admet_id_to_units() -> dict[str, str]:
    """Get the ADMET ID to units mapping.

    :return: A dictionary mapping ADMET IDs to units.
    """
    return ADMET_ID_TO_UNITS



"""Runs the production web interface for ADMET-AI (Gunicorn)."""
from flask import Flask

from admet_ai.web.app import app
from admet_ai.web.run import setup_web


def build_app(*args, **kwargs) -> Flask:
    setup_web(**kwargs)

    return app



"""Defines functions for the DrugBank approved reference set."""
from collections import defaultdict
from functools import lru_cache
from pathlib import Path

import pandas as pd

from admet_ai.admet_info import get_admet_id_to_name
from admet_ai.constants import (
    DEFAULT_DRUGBANK_PATH,
    DRUGBANK_ATC_NAME_PREFIX,
    DRUGBANK_ATC_PREFIX,
    DRUGBANK_DELIMITER,
    DRUGBANK_ID_COLUMN,
    DRUGBANK_NAME_COLUMN,
    DRUGBANK_SMILES_COLUMN,
)


DRUGBANK_DF = pd.DataFrame()
ATC_CODE_TO_DRUGBANK_INDICES: dict[str, list[int]] = {}


def load_drugbank(drugbank_path: Path = DEFAULT_DRUGBANK_PATH) -> None:
    """Loads the reference set of DrugBank approved molecules with their model predictions.

    :param drugbank_path: The path to the DrugBank reference set.
    """
    # Set up global variables
    global DRUGBANK_DF, ATC_CODE_TO_DRUGBANK_INDICES

    # Load DrugBank DataFrame
    DRUGBANK_DF = pd.read_csv(drugbank_path)

    # Map ATC codes to all indices of the DRUGBANK_DF with that ATC code
    atc_code_to_drugbank_indices = defaultdict(set)
    for atc_column in [
        column
        for column in DRUGBANK_DF.columns
        if column.startswith(DRUGBANK_ATC_NAME_PREFIX)
    ]:
        for index, atc_codes in DRUGBANK_DF[atc_column].dropna().items():
            for atc_code in atc_codes.split(DRUGBANK_DELIMITER):
                atc_code_to_drugbank_indices[atc_code.lower()].add(index)

    # Save ATC code to indices mapping to global variable and convert set to sorted list
    ATC_CODE_TO_DRUGBANK_INDICES = {
        atc_code: sorted(indices)
        for atc_code, indices in atc_code_to_drugbank_indices.items()
    }


def get_drugbank(atc_code: str | None = None) -> pd.DataFrame:
    """Get the DrugBank reference DataFrame, optionally filtered by ATC code.

    :param atc_code: The ATC code to filter by. If None or 'all', returns the entire DrugBank.
    :return: A DataFrame containing the DrugBank reference set, optionally filtered by ATC code.
    """
    if DRUGBANK_DF.empty:
        load_drugbank()

    if atc_code is None:
        return DRUGBANK_DF

    return DRUGBANK_DF.loc[ATC_CODE_TO_DRUGBANK_INDICES[atc_code]]


@lru_cache()
def get_drugbank_size(atc_code: str | None = None) -> int:
    """Get the number of molecules in the DrugBank reference set, optionally filtered by ATC code.

    :param atc_code: The ATC code to filter by. If None or 'all', returns the entire DrugBank.
    :return: The number of molecules in the DrugBank reference set, optionally filtered by ATC code.
    """
    # Get DrugBank reference, optionally filtered ATC code
    drugbank = get_drugbank(atc_code=atc_code)

    return len(drugbank)


@lru_cache()
def get_drugbank_unique_atc_codes() -> list[str]:
    """Get the unique ATC codes in the DrugBank reference set.

    :return: A list of unique ATC codes in the DrugBank reference set.
    """
    drugbank = get_drugbank()

    return sorted(
        {
            atc_code.lower()
            for atc_column in [
                column
                for column in drugbank.columns
                if column.startswith(DRUGBANK_ATC_NAME_PREFIX)
            ]
            for atc_codes in drugbank[atc_column].dropna().str.split(DRUGBANK_DELIMITER)
            for atc_code in atc_codes
        }
    )


@lru_cache()
def get_drugbank_tasks_ids() -> list[str]:
    """Get the tasks (properties) predicted by the DrugBank reference set.

    :return: A list of tasks (properties) predicted in the DrugBank reference set.
    """
    drugbank = get_drugbank()

    non_task_columns = [
        DRUGBANK_ID_COLUMN,
        DRUGBANK_NAME_COLUMN,
        DRUGBANK_SMILES_COLUMN,
    ] + [
        column for column in drugbank.columns if column.startswith(DRUGBANK_ATC_PREFIX)
    ]
    task_columns = set(drugbank.columns) - set(non_task_columns)
    drugbank_task_ids = sorted(task_columns)

    return drugbank_task_ids


@lru_cache()
def get_drugbank_task_names() -> list[str]:
    """Get the names of the tasks (properties) predicted by the DrugBank reference set.

    :return: A list of task names (properties) predicted in the DrugBank reference set.
    """
    admet_id_to_name = get_admet_id_to_name()
    drugbank_task_names = sorted(
        [admet_id_to_name[task_id] for task_id in get_drugbank_tasks_ids()],
        key=lambda name: name.lower(),
    )

    return drugbank_task_names



"""Defines the routes of the ADMET-AI Flask app."""
from uuid import uuid4
from tempfile import NamedTemporaryFile

from flask import (
    after_this_request,
    jsonify,
    render_template,
    request,
    Response,
    send_file,
    session,
)

from admet_ai._version import __version__
from admet_ai.admet_info import get_admet_info
from admet_ai.drugbank import (
    get_drugbank,
    get_drugbank_size,
    get_drugbank_task_names,
    get_drugbank_unique_atc_codes,
)
from admet_ai.plot import (
    plot_drugbank_reference,
    plot_molecule_svg,
    plot_radial_summary,
)
from admet_ai.utils import get_drugbank_suffix
from admet_ai.web.app import app
from admet_ai.web.app.models import get_admet_model
from admet_ai.web.app.storage import (
    get_user_preds,
    set_user_preds,
    update_user_activity,
)
from admet_ai.web.app.utils import (
    get_smiles_from_request,
    replace_svg_dimensions,
    smiles_to_mols,
    string_to_html_sup,
)


def render(**kwargs) -> str:
    """Renders the page with specified kwargs."""
    return render_template(
        "index.html",
        admet_info=get_admet_info(),
        drugbank_atc_codes=["all"] + get_drugbank_unique_atc_codes(),
        drugbank_tasks=get_drugbank_task_names(),
        max_molecules=app.config["MAX_MOLECULES"],
        max_visible_molecules=app.config["MAX_VISIBLE_MOLECULES"],
        low_performance_threshold=app.config["LOW_PERFORMANCE_THRESHOLD"],
        drugbank_approved_percentile_suffix=get_drugbank_suffix(
            session.get("atc_code")
        ),
        drugbank_size=get_drugbank_size(session.get("atc_code")),
        drugbank_total_size=get_drugbank_size(),
        atc_code=session.get("atc_code") or "all",
        heartbeat_frequency=app.config["HEARTBEAT_FREQUENCY"],
        string_to_html_sup=string_to_html_sup,
        min=min,
        max=max,
        text_smiles=session.get("text_smiles", ""),
        version=__version__,
        **kwargs,
    )


@app.route("/", methods=["GET", "POST"])
def index() -> str:
    """Renders the page and makes predictions if the method is POST."""
    # Set up warnings
    warnings = []

    # Assign user ID to session
    if "user_id" not in session:
        session["user_id"] = uuid4().hex

    # If GET request, simply return the page; otherwise if POST request, make predictions
    if request.method == "GET":
        return render()

    # Save the user's text box input
    session["text_smiles"] = request.form["text-smiles"]

    # Get the SMILES from the request
    all_smiles, error = get_smiles_from_request()

    # Return any errors
    if error is not None:
        return render(errors=[error])

    # Error if too many molecules
    if (
        app.config["MAX_MOLECULES"] is not None
        and len(all_smiles) > app.config["MAX_MOLECULES"]
    ):
        return render(
            errors=[
                f"Received too many molecules. Maximum number of molecules is {app.config['MAX_MOLECULES']:,}."
            ]
        )

    # Convert SMILES to RDKit molecules
    mols = smiles_to_mols(all_smiles)

    # Warn about invalid SMILES
    for smiles, mol in zip(all_smiles, mols):
        if mol is None or " " in smiles:
            warnings.append(f"Invalid SMILES string: {smiles}")

    # Remove invalid molecules
    all_smiles = [smile for smile, mol in zip(all_smiles, mols) if mol is not None]
    mols = [mol for mol in mols if mol is not None]

    # Error if no valid molecules
    if len(all_smiles) == 0:
        return render(errors=["No valid SMILES strings given."])

    # Make physicochemical and ADMET predictions (with DrugBank percentiles)
    admet_model = get_admet_model()
    admet_model.atc_code = session.get("atc_code")
    all_preds = admet_model.predict(smiles=all_smiles)

    # Convert predictions to a dictionary mapping SMILES to property name to value
    smiles_to_property_id_to_pred: dict[str, dict[str, float]] = all_preds[
        ~all_preds.index.duplicated(keep="first")  # drop duplicate SMILES indices
    ].to_dict(orient="index")

    # Store predictions in memory
    set_user_preds(user_id=session["user_id"], preds_df=all_preds)

    # Create DrugBank reference plot
    drugbank_plot_svg = plot_drugbank_reference(
        preds_df=all_preds,
        drugbank_df=get_drugbank(atc_code=session.get("atc_code")),
        x_property_name=session.get("drugbank_x_task_name"),
        y_property_name=session.get("drugbank_y_task_name"),
        max_molecule_num=app.config["MAX_VISIBLE_MOLECULES"],
    ).decode("utf-8")
    drugbank_plot_svg = replace_svg_dimensions(drugbank_plot_svg)

    # Get maximum number of molecules to display
    num_display_molecules = min(len(all_smiles), app.config["MAX_VISIBLE_MOLECULES"])

    # Create molecule SVG images
    mol_svgs = [plot_molecule_svg(mol) for mol in mols[:num_display_molecules]]
    mol_svgs = [replace_svg_dimensions(plot) for plot in mol_svgs]

    # Create molecule radial plots for DrugBank approved percentiles
    radial_svgs = [
        plot_radial_summary(
            property_id_to_percentile=smiles_to_property_id_to_pred[smiles],
            percentile_suffix=get_drugbank_suffix(session.get("atc_code")),
        ).decode("utf-8")
        for smiles in all_smiles[:num_display_molecules]
    ]
    radial_svgs = [replace_svg_dimensions(plot) for plot in radial_svgs]

    return render(
        predicted=True,
        all_smiles=all_smiles,
        smiles_to_property_id_to_pred=smiles_to_property_id_to_pred,
        mol_svgs=mol_svgs,
        radial_svgs=radial_svgs,
        drugbank_plot=drugbank_plot_svg,
        num_molecules=len(all_smiles),
        num_display_molecules=num_display_molecules,
        warnings=warnings,
    )


@app.route("/set_atc_code", methods=["POST"])
def set_atc_code() -> Response:
    """Sets the ATC code to filter the DrugBank reference set by.

    :return: A JSON response containing the new DrugBank size.
    """
    # Get ATC code
    atc_code = request.args.get("atc_code")

    # Handle "all" ATC code
    if atc_code == "all":
        atc_code = None

    # Store ATC code in session
    session["atc_code"] = atc_code

    # Get new DrugBank size
    drugbank_size = get_drugbank_size(session.get("atc_code"))

    # Send new DrugBank size
    return jsonify({"drugbank_size_string": f"{drugbank_size:,}",})


@app.route("/drugbank_plot", methods=["GET"])
def drugbank_plot() -> Response:
    """Creates a DrugBank reference plot.

    :return: A JSON response containing the SVG of the plot.
    """
    # Get requested X and Y axes and store in session
    session["drugbank_x_task_name"] = request.args.get(
        "x_task", default=session.get("drugbank_x_task_name"), type=str
    )
    session["drugbank_y_task_name"] = request.args.get(
        "y_task", default=session.get("drugbank_y_task_name"), type=str
    )

    # Create DrugBank reference plot with ATC code
    drugbank_plot_svg = plot_drugbank_reference(
        preds_df=get_user_preds(session["user_id"]),
        drugbank_df=get_drugbank(atc_code=session.get("atc_code")),
        x_property_name=session["drugbank_x_task_name"],
        y_property_name=session["drugbank_y_task_name"],
        max_molecule_num=app.config["MAX_VISIBLE_MOLECULES"],
    ).decode("utf-8")
    drugbank_plot_svg = replace_svg_dimensions(drugbank_plot_svg)

    return jsonify({"svg": drugbank_plot_svg})


@app.route("/download_predictions")
def download_predictions() -> Response:
    """Downloads predictions as a CSV file.

    :return: A CSV file containing the predictions.
    """
    # Create a temporary file to hold the predictions
    preds_file = NamedTemporaryFile()

    # Set up a function to close the file after the response is sent
    @after_this_request
    def remove_file(response: Response) -> Response:
        preds_file.close()
        return response

    # Save predictions to temporary file
    get_user_preds(session["user_id"]).to_csv(preds_file.name, index_label="smiles")
    preds_file.seek(0)

    # Return the temporary file as a response
    return send_file(
        preds_file.name, as_attachment=True, download_name="predictions.csv"
    )


@app.route("/heartbeat", methods=["POST"])
def heartbeat() -> tuple[str, int]:
    """Registers that the client is still using the site.

    :return: A tuple containing an empty string and a 204 status code.
    """
    # Update user's last activity
    session.modified = True

    if "user_id" in session:
        update_user_activity(session["user_id"])

    # Send no content response
    return "", 204



"""Sets the config parameters for the ADMET-AI Flask app object."""
from admet_ai.constants import (
    DEFAULT_DRUGBANK_PATH,
    DEFAULT_MODELS_DIR,
)


MODELS_DIR = DEFAULT_MODELS_DIR
DRUGBANK_PATH = DEFAULT_DRUGBANK_PATH
LOW_PERFORMANCE_THRESHOLD = 0.6
NUM_WORKERS = 0



from pathlib import Path
from setuptools import find_packages, setup

# Load version number
__version__ = ""
version_file = Path(__file__).parent.absolute() / "admet_ai" / "_version.py"

with open(version_file) as fd:
    exec(fd.read())

# Load README
with open("README.md", encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="admet_ai",
    version=__version__,
    author="Kyle Swanson",
    author_email="swansonk.14@gmail.com",
    description="admet_ai",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/swansonk14/admet_ai",
    download_url=f"https://github.com/swansonk14/admet_ai/archive/refs/tags/v_{__version__}.tar.gz",
    license="MIT",
    packages=find_packages(),
    package_data={"admet_ai": ["py.typed", "resources/**/*"]},
    entry_points={
        "console_scripts": [
            "admet_predict=admet_ai.admet_predict:admet_predict_command_line",
            "admet_web=admet_ai.web.run:admet_web",
        ]
    },
    install_requires=[
        "chemfunc>=1.0.4",
        "chemprop==1.6.1",
        "numpy",
        "pandas>=2.0.0,<2.2.0",  # remove this limit once rdkit implements a fix to PandasTools
        "rdkit>=2023.3.3",
        "seaborn",
        "tqdm",
        "typed-argument-parser>=1.9.0",
    ],
    extras_require={
        "tdc": ["openpyxl", "PyTDC>=0.4.1"],
        "web": ["flask", "gunicorn"],
    },
    python_requires=">=3.10",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Typing :: Typed",
    ],
    keywords=[
        "machine learning",
        "drug design",
        "ADMET",
        "molecular property prediction",
    ],
)



"""ADMET-AI class to contain ADMET model and prediction function."""
from collections import defaultdict
from multiprocessing import Pool
from pathlib import Path

import numpy as np
import pandas as pd
import torch
from chemfunc.molecular_fingerprints import compute_rdkit_fingerprint
from chemprop.data import (
    MoleculeDataLoader,
    MoleculeDatapoint,
    MoleculeDataset,
    set_cache_graph,
    set_cache_mol,
)
from chemprop.data.data import SMILES_TO_MOL
from chemprop.models import MoleculeModel
from chemprop.train import predict
from chemprop.utils import load_args, load_checkpoint, load_scalers
from rdkit import Chem
from scipy.stats import percentileofscore
from sklearn.preprocessing import StandardScaler
from tqdm import tqdm

from admet_ai.constants import (
    DEFAULT_DRUGBANK_PATH,
    DEFAULT_MODELS_DIR,
    DRUGBANK_ATC_NAME_PREFIX,
    DRUGBANK_DELIMITER,
)
from admet_ai.physchem import compute_physicochemical_properties


class ADMETModel:
    """ADMET-AI model class."""

    def __init__(
        self,
        models_dir: Path | str = DEFAULT_MODELS_DIR,
        include_physchem: bool = True,
        drugbank_path: Path | str | None = DEFAULT_DRUGBANK_PATH,
        atc_code: str | None = None,
        num_workers: int | None = None,
        cache_molecules: bool = True,
        fingerprint_multiprocessing_min: int = 100,
    ) -> None:
        """Initialize the ADMET-AI model.

        :param models_dir: Path to a directory containing subdirectories, each of which contains an ensemble
                           of Chemprop-RDKit models.
        :param include_physchem: Whether to include physicochemical properties in the predictions.
        :param drugbank_path: Path to a CSV file containing DrugBank approved molecules
                              with ADMET predictions and ATC codes.
        :param atc_code: The ATC code to filter the DrugBank reference set by.
                         If None, the entire DrugBank reference set will be used.
        :param num_workers: Number of workers for the data loader. Zero workers (i.e., sequential data loading)
                            may be faster if not using a GPU, while multiple workers (e.g., 8) are faster with a GPU.
                            If None, defaults to 0 if no GPU is available and 8 if a GPU is available.
        :param cache_molecules: Whether to cache molecules. Caching improves prediction speed but requires more memory.
        :param fingerprint_multiprocessing_min: Minimum number of molecules for multiprocessing to be used for
                                                fingerprint computation. Otherwise, single processing is used.
        """
        # Check parameters
        if atc_code is not None and drugbank_path is None:
            raise ValueError(
                "DrugBank reference set must be provided to filter by ATC code."
            )

        # Set default num_workers
        if num_workers is None:
            num_workers = 8 if torch.cuda.is_available() else 0

        # Save parameters
        self.include_physchem = include_physchem
        self.num_workers = num_workers
        self.cache_molecules = cache_molecules
        self.fingerprint_multiprocessing_min = fingerprint_multiprocessing_min
        self._atc_code = atc_code

        # Load DrugBank reference set if needed
        if drugbank_path is not None:
            # Load DrugBank DataFrame
            self.drugbank = pd.read_csv(drugbank_path)

            # Map ATC codes to all indices of the drugbank with that ATC code
            atc_code_to_drugbank_indices = defaultdict(set)
            for atc_column in [
                column for column in self.drugbank.columns if column.startswith(DRUGBANK_ATC_NAME_PREFIX)
            ]:
                for index, atc_codes in self.drugbank[atc_column].dropna().items():
                    for atc_code in atc_codes.split(DRUGBANK_DELIMITER):
                        atc_code_to_drugbank_indices[atc_code.lower()].add(index)

            # Save ATC code to indices mapping to global variable and convert set to sorted list
            self.atc_code_to_drugbank_indices = {
                atc_code: sorted(indices)
                for atc_code, indices in atc_code_to_drugbank_indices.items()
            }

            # Filter DrugBank by ATC code if needed
            if self.atc_code is not None:
                self.drugbank_atc_filtered = self.drugbank.loc[
                    self.atc_code_to_drugbank_indices[self.atc_code]
                ]
            else:
                self.drugbank_atc_filtered = self.drugbank
        else:
            self.drugbank = self.drugbank_atc_filtered = None

        # Set caching
        set_cache_graph(self.cache_molecules)
        set_cache_mol(self.cache_molecules)

        # Set device based on GPU availability
        self.device = (
            torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")
        )

        # Prepare lists to contain model details
        self.task_lists: list[list[str]] = []
        self.use_features_list: list[bool] = []
        self.model_lists: list[list[MoleculeModel]] = []
        self.scaler_lists: list[list[StandardScaler | None]] = []

        # Get model ensemble directories
        model_dirs = sorted(models_dir.iterdir())

        # Load each ensemble of models
        for model_dir in model_dirs:
            # Get model paths for the ensemble in the directory
            model_paths = sorted(Path(model_dir).glob("**/*.pt"))

            # Load args for this ensemble
            train_args = load_args(str(model_paths[0]))

            # Get task names for this ensemble
            task_names = train_args.task_names
            self.task_lists.append(task_names)

            # Get whether to use features for this ensemble
            use_features = train_args.use_input_features
            self.use_features_list.append(use_features)

            # Load models in the ensemble
            models = [
                load_checkpoint(path=str(model_path), device=self.device).eval()
                for model_path in model_paths
            ]
            self.model_lists.append(models)

            # Load scalers for each model
            scalers = [
                load_scalers(path=str(model_path))[0] for model_path in model_paths
            ]
            self.scaler_lists.append(scalers)

        # Ensure all models do or do not use features
        if not len(set(self.use_features_list)) == 1:
            raise ValueError("All models must either use or not use features.")

        self.use_features = self.use_features_list[0]

    @property
    def num_ensembles(self) -> int:
        """Get the number of ensembles."""
        return len(self.model_lists)

    @property
    def atc_code(self) -> str | None:
        """Get the ATC code."""
        return self._atc_code

    @atc_code.setter
    def atc_code(self, atc_code: str | None) -> None:
        """Set the ATC code and filter DrugBank based on provided ATC code.

        :param atc_code: The ATC code to filter the DrugBank reference set by.
                         If None, the entire DrugBank reference set will be used.
        """
        # Handle case of no DrugBank
        if self.drugbank is None:
            raise ValueError(
                "Cannot set ATC code if DrugBank reference is not provided."
            )

        # Validate ATC code
        if atc_code is not None and atc_code not in self.atc_code_to_drugbank_indices:
            raise ValueError(f"Invalid ATC code: {atc_code}")

        # Save ATC code
        self._atc_code = atc_code

        # Filter DrugBank by ATC code if needed
        if self.atc_code is not None:
            self.drugbank_atc_filtered = self.drugbank.loc[
                self.atc_code_to_drugbank_indices[self.atc_code]
            ]
        else:
            self.drugbank_atc_filtered = self.drugbank

    def predict(self, smiles: str | list[str]) -> dict[str, float] | pd.DataFrame:
        """Make predictions on a list of SMILES strings.

        :param smiles: A SMILES string or a list of SMILES strings.
        :return: If smiles is a string, returns a dictionary mapping property name to prediction.
                 If smiles is a list, returns a DataFrame containing the predictions with SMILES strings as the index
                 and property names as the columns.
        """
        # Convert SMILES to list if needed
        if isinstance(smiles, str):
            smiles = [smiles]
            smiles_type = str
        else:
            smiles_type = list

        # Convert SMILES to RDKit molecules and cache if desired
        mols = []
        for smile in tqdm(smiles, desc="SMILES to Mol"):
            if smile in SMILES_TO_MOL:
                mol = SMILES_TO_MOL[smile]
            else:
                mol = Chem.MolFromSmiles(smile)

            mols.append(mol)

            if self.cache_molecules:
                SMILES_TO_MOL[smile] = mol

        # Remove invalid molecules
        invalid_mols = [mol is None for mol in mols]

        if any(invalid_mols):
            print(f"Warning: {sum(invalid_mols):,} invalid molecules will be removed")

            mols = [mol for mol in mols if mol is not None]
            smiles = [
                smile for smile, invalid in zip(smiles, invalid_mols) if not invalid
            ]

        # Compute physicochemical properties
        physchem_preds = compute_physicochemical_properties(
            all_smiles=smiles, mols=mols
        )

        # Compute fingerprints if needed
        if self.use_features:
            # Select between multiprocessing and single processing
            if len(mols) >= self.fingerprint_multiprocessing_min:
                pool = Pool()
                map_fn = pool.imap
            else:
                pool = None
                map_fn = map

            # Compute fingerprints
            fingerprints = np.array(
                list(
                    tqdm(
                        map_fn(compute_rdkit_fingerprint, mols),
                        total=len(mols),
                        desc=f"RDKit fingerprints",
                    )
                )
            )

            # Close pool if needed
            if pool is not None:
                pool.close()
        else:
            fingerprints = [None] * len(smiles)

        # Build data loader
        data_loader = MoleculeDataLoader(
            dataset=MoleculeDataset(
                [
                    MoleculeDatapoint(smiles=[smile], features=fingerprint,)
                    for smile, fingerprint in zip(smiles, fingerprints)
                ]
            ),
            num_workers=self.num_workers,
            shuffle=False,
        )

        # Make predictions
        task_to_preds = {}

        # Loop through each ensemble and make predictions
        for tasks, use_features, models, scalers in tqdm(
            zip(
                self.task_lists,
                self.use_features_list,
                self.model_lists,
                self.scaler_lists,
            ),
            total=self.num_ensembles,
            desc="model ensembles",
        ):
            # Make predictions
            preds = [
                predict(model=model, data_loader=data_loader)
                for model in tqdm(models, desc="individual models")
            ]

            # Scale predictions if needed (for regression)
            if scalers[0] is not None:
                preds = [
                    scaler.inverse_transform(pred).astype(float)
                    for scaler, pred in zip(scalers, preds)
                ]

            # Average ensemble predictions
            preds = np.mean(preds, axis=0)

            # Add predictions to data
            for i, task in enumerate(tasks):
                task_to_preds[task] = preds[:, i]

        # Put preds in a DataFrame
        admet_preds = pd.DataFrame(task_to_preds, index=smiles)

        # Combine physicochemical and ADMET properties
        preds = pd.concat((physchem_preds, admet_preds), axis=1)

        # Compute DrugBank percentiles if needed
        if self.drugbank is not None:
            # Set DrugBank suffix
            if self.atc_code is None:
                drugbank_suffix = "drugbank_approved_percentile"
            else:
                drugbank_suffix = f"drugbank_approved_{self.atc_code}_percentile"

            # Compute DrugBank percentiles
            drugbank_percentiles = pd.DataFrame(
                data={
                    f"{property_name}_{drugbank_suffix}": percentileofscore(
                        self.drugbank_atc_filtered[property_name],
                        preds[property_name].values,
                    )
                    for property_name in preds.columns
                },
                index=smiles,
            )

            # Combine predictions and percentiles
            preds = pd.concat((preds, drugbank_percentiles), axis=1)

        # Convert to dictionary if SMILES type is string
        if smiles_type == str:
            preds = preds.iloc[0].to_dict()

        return preds



"""Utility functions for ADMET-AI."""
from pathlib import Path

import pandas as pd


def load_and_preprocess_data(
    data_path: Path, smiles_column: str = "smiles"
) -> pd.DataFrame:
    """Preprocess a dataset of molecules by removing missing SMILES and setting the SMILES as the index.

    :param data_path: Path to a CSV file containing a dataset of molecules.
    :param smiles_column: Name of the column containing SMILES strings.
    :return: A DataFrame containing the preprocessed data with SMILES strings as the index.
    """
    # Load data
    data = pd.read_csv(data_path)

    # Remove missing SMILES
    original_num_molecules = len(data)
    data.dropna(subset=[smiles_column], inplace=True, ignore_index=True)

    # Warn if molecules were removed
    if len(data) < original_num_molecules:
        print(
            f"Warning: {original_num_molecules - len(data):,} molecules were removed "
            f"from the dataset because they were missing SMILES."
        )

    # Set SMILES as index
    data.set_index(smiles_column, inplace=True)

    return data


def get_drugbank_suffix(atc_code: str | None) -> str:
    """Gets the DrugBank percentile suffix for the given ATC code.

    :param atc_code: The ATC code.
    """
    if atc_code is None:
        return "drugbank_approved_percentile"

    return f"drugbank_approved_{atc_code}_percentile"
